""" Url router for the core elasticsearch app
"""
from django.urls import re_path

from core_elasticsearch_app.rest.elasticsearch_template.views import (
    ElasticsearchTemplateList,
)
from core_elasticsearch_app.rest.views import (
    DocumentSuggestion,
    KeywordSuggestion,
)

urlpatterns = [
    re_path(
        r"^elasticsearch_template/$",
        ElasticsearchTemplateList.as_view(),
        name="core_elasticsearch_app_rest_elasticsearch_template_list",
    ),
    re_path(
        r"^document/suggest",
        DocumentSuggestion.as_view(),
        name="core_elasticsearch_document_suggestion_view",
    ),
    re_path(
        r"^keyword/suggest",
        KeywordSuggestion.as_view(),
        name="core_elasticsearch_keyword_suggestion_view",
    ),
]



""" Url router for the administration site
"""
from django.contrib import admin
from django.contrib.admin.views.decorators import staff_member_required
from django.urls import re_path

from core_main_app.admin import core_admin_site
from core_elasticsearch_app.components.elasticsearch_template.models import (
    ElasticsearchTemplate,
)
from core_elasticsearch_app.views.admin import (
    views as admin_views,
    ajax as admin_ajax,
)


admin_urls = [
    re_path(
        r"^elasticsearch/templates",
        staff_member_required(admin_views.TemplatesView.as_view()),
        name="core_elasticsearch_app_templates",
    ),
    re_path(
        r"^elasticsearch/template/add",
        staff_member_required(
            admin_ajax.AddElasticsearchTemplateView.as_view()
        ),
        name="core_elasticsearch_app_templates_add",
    ),
    re_path(
        r"^elasticsearch/template/(?P<pk>[\w-]+)/delete/$",
        staff_member_required(
            admin_ajax.DeleteElasticsearchTemplateView.as_view()
        ),
        name="core_elasticsearch_app_templates_delete",
    ),
    re_path(
        r"^elasticsearch/template/(?P<pk>[\w-]+)/edit/$",
        staff_member_required(
            admin_ajax.EditElasticsearchTemplateView.as_view()
        ),
        name="core_elasticsearch_app_templates_edit",
    ),
    re_path(
        r"^elasticsearch/template/(?P<pk>[\w-]+)/check/$",
        admin_ajax.check_data_from_template,
        name="core_elasticsearch_app_templates_check",
    ),
    re_path(
        r"^elasticsearch/template/(?P<pk>[\w-]+)/index/$",
        admin_ajax.index_data_from_template,
        name="core_elasticsearch_app_templates_index",
    ),
]

admin.site.register(ElasticsearchTemplate)
urls = core_admin_site.get_urls()
core_admin_site.get_urls = lambda: admin_urls + urls



""" ElasticsearchTemplate api
"""
from core_elasticsearch_app.components.elasticsearch_template.models import (
    ElasticsearchTemplate,
)


def upsert(elasticsearch_template):
    """Save or Update Elasticsearch Template

    Args:
        elasticsearch_template:

    Returns:

    """
    elasticsearch_template.save()


def delete(elasticsearch_template):
    """Delete Elasticsearch Template

    Args:
        elasticsearch_template:

    Returns:

    """
    elasticsearch_template.delete()


def get_by_template(template):
    """Get Elasticsearch Template

    Args:
        template:

    Returns:

    """
    return ElasticsearchTemplate.get_by_template(template)


def get_all():
    """Get all Elasticsearch Template

    Returns:

    """
    return ElasticsearchTemplate.get_all()


def get_by_id(es_template_id):
    """Get Elasticsearch Template by id

    Args:
        es_template_id:


    Returns:

    """
    return ElasticsearchTemplate.get_by_id(es_template_id)



""" Apps file for setting elasticsearch when app is ready
"""
import sys

from django.apps import AppConfig

from core_elasticsearch_app.commons.exceptions import ElasticsearchError
from core_elasticsearch_app.components.data.elasticsearch import (
    create_data_index,
)
from core_elasticsearch_app.settings import (
    CAN_SET_PUBLIC_DATA_TO_PRIVATE,
    CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT,
)


class ElasticsearchAppConfig(AppConfig):
    """Core application settings"""

    name = "core_elasticsearch_app"

    def ready(self):
        """Run when the app is ready

        Returns:

        """
        if "migrate" not in sys.argv:
            from core_elasticsearch_app.components.data import (
                watch as data_watch,
            )

            if (
                CAN_SET_PUBLIC_DATA_TO_PRIVATE
                and not CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT
            ):
                raise ElasticsearchError(
                    "The Elasticsearch app will only work for systems where "
                    "CAN_SET_PUBLIC_DATA_TO_PRIVATE is set to False,"
                    " and CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT is set to True."
                )
            create_data_index()
            data_watch.init()



""" Autocomplete settings for the data collection
"""

# settings for CDCS data
DATA_AUTOCOMPLETE_SETTINGS = {
    "settings": {
        "analysis": {
            "filter": {
                "autocomplete_filter": {
                    "type": "edge_ngram",
                    "min_gram": 1,
                    "max_gram": 20,
                }
            },
            "analyzer": {
                "autocomplete": {
                    "type": "custom",
                    "tokenizer": "standard",
                    "filter": ["lowercase", "autocomplete_filter"],
                }
            },
        }
    },
    "mappings": {
        "properties": {
            "title": {
                "type": "text",
                "analyzer": "autocomplete",
                "search_analyzer": "standard",
            },
            "description": {
                "type": "text",
                "analyzer": "autocomplete",
                "search_analyzer": "standard",
            },
        }
    },
}



""" Serializers used for the elasticsearch template REST API.
"""
from rest_framework.serializers import ModelSerializer

from core_elasticsearch_app.components.elasticsearch_template import (
    api as elasticsearch_template_api,
)
from core_elasticsearch_app.components.elasticsearch_template.models import (
    ElasticsearchTemplate,
)


class ElasticsearchTemplateSerializer(ModelSerializer):
    """ElasticsearchTemplate serializer"""

    class Meta:
        """Meta"""

        model = ElasticsearchTemplate
        fields = ["id", "template", "title_path", "description_paths"]
        read_only_fields = ("id",)

    def create(self, validated_data):
        """Create and return a new `ElasticsearchTemplate` instance, given the validated data."""
        # Create instance from the validated data and insert it in DB
        instance = ElasticsearchTemplate(
            template=validated_data["template"],
            title_path=validated_data["title_path"],
            description_paths=validated_data["description_paths"],
        )
        elasticsearch_template_api.upsert(instance)

        return instance



""" Utils for managing CDCS data in elasticsearch
"""
import logging

from core_main_app.commons import exceptions
from xml_utils.xsd_tree.xsd_tree import XSDTree

from core_elasticsearch_app.components.data.autocomplete_settings import (
    DATA_AUTOCOMPLETE_SETTINGS,
)
from core_elasticsearch_app.components.data.mongodb import get_value_from_path
from core_elasticsearch_app.settings import (
    ELASTICSEARCH_CDCS_DATA_INDEX,
)
from core_elasticsearch_app.utils.elasticsearch_client import (
    ElasticsearchClient,
)
from core_elasticsearch_app.components.elasticsearch_template import (
    api as elasticsearch_template_api,
)

logger = logging.getLogger(__name__)


def create_data_index():
    """Create autocomplete index on data titles

    Returns:

    """
    return ElasticsearchClient.create_index(
        ELASTICSEARCH_CDCS_DATA_INDEX, DATA_AUTOCOMPLETE_SETTINGS
    )


def index_data(data):
    """Build json document for elasticsearch from CDCS data

    Args:
        data:

    Returns:

    """

    try:
        es_template = elasticsearch_template_api.get_by_template(data.template)
        es_data = {
            "data_id": str(data.id),
            "title": get_value_from_path(data, es_template.title_path)
            if es_template.title_path
            else data.title,
            "description": " ".join(
                [
                    _get_string_value(get_value_from_path(data, path))
                    for path in es_template.description_paths
                ]
            )
            if es_template.description_paths
            else XSDTree.fromstring(data.xml_content).xpath("//text()"),
        }
        # TODO: could use global PID instead of id
        return ElasticsearchClient.index_document(
            ELASTICSEARCH_CDCS_DATA_INDEX, str(data.id), es_data
        )
    except exceptions.DoesNotExist:
        logger.warning(
            "Data with id %s will not be indexed. No template configured.",
            str(data.id),
        )
    except Exception as exception:
        logger.error(str(exception))


def get_suggestions(query, fuzziness=1, prefix_length=3, fragment_size=100):
    """Get suggestions from query

    Args:
        query:
        fuzziness:
        prefix_length:
        fragment_size:

    Returns:

    """
    query = {
        "query": {
            "multi_match": {
                "query": query,
                "fuzziness": fuzziness,
                "prefix_length": prefix_length,
            }
        },
        "highlight": {
            "number_of_fragments": 1,
            "fragment_size": fragment_size,
            "fields": {"title": {}, "description": {}},
        },
    }
    result = ElasticsearchClient.search(ELASTICSEARCH_CDCS_DATA_INDEX, query)
    return result["hits"]["hits"]


def _get_string_value(value):
    """Get string value

    Args:
        value:

    Returns:

    """
    if value:
        return value
    return ""



""" Elasticsearch menu
"""
from django.urls import reverse
from menu import Menu, MenuItem


elasticsearch_children = (
    MenuItem(
        "Configure Templates",
        reverse("core-admin:core_elasticsearch_app_templates"),
        icon="list",
    ),
)

Menu.add_item(
    "admin", MenuItem("ELASTICSEARCH", None, children=elasticsearch_children)
)



""" Tests Settings
"""

SECRET_KEY = "fake-key"

INSTALLED_APPS = [
    # Django apps
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.sites",
    # Local apps
    "tests",
]

MIDDLEWARE = (
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": ["templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.i18n",
            ],
        },
    },
]
MONGODB_INDEXING = False
MONGODB_ASYNC_SAVE = False



""" Elasticsearch tasks
"""
import logging

from celery import shared_task

from core_main_app.system import api as data_system_api
from core_elasticsearch_app.components.data import elasticsearch

logger = logging.getLogger(__name__)


@shared_task
def index_all_data_from_template(template):
    """Index all data"""
    try:
        data = data_system_api.get_all_by_template(template=template)
        for document in data:
            try:
                elasticsearch.index_data(document)
            except Exception as exception:
                logger.error(
                    "ERROR : An error occurred while indexing data : %s",
                    str(exception),
                )
    except Exception as exception:
        logger.error(
            "ERROR : An error occurred while indexing data : %s",
            str(exception),
        )


@shared_task
def index_data(data_id):
    """Index a data"""
    try:
        data = data_system_api.get_data_by_id(data_id)
        try:
            elasticsearch.index_data(data)
        except Exception as exception:
            logger.error(
                "ERROR : An error occurred while indexing data :  %s",
                str(exception),
            )
    except Exception as exception:
        logger.error(
            "ERROR : An error occurred while indexing data :  %s",
            str(exception),
        )



""" REST views for the elasticsearch template.
"""
from django.utils.decorators import method_decorator
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView

from core_main_app.commons import exceptions
from core_main_app.utils.decorators import api_staff_member_required
from core_elasticsearch_app.components.elasticsearch_template import (
    api as elasticsearch_template_api,
)
from core_elasticsearch_app.rest.elasticsearch_template.serializers import (
    ElasticsearchTemplateSerializer,
)


class ElasticsearchTemplateList(APIView):
    """List elasticsearch template configuration"""

    @method_decorator(api_staff_member_required())
    def get(self, request):
        """Get all elasticsearch template configuration

        Args:
            request: HTTP request

        Returns:

            - code: 200
              content: List of elasticsearch template configurations
            - code: 500
              content: Internal server error
        """
        try:
            object_list = elasticsearch_template_api.get_all()

            # Serialize object
            serializer = ElasticsearchTemplateSerializer(
                object_list, many=True
            )

            # Return response
            return Response(serializer.data, status=status.HTTP_200_OK)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(
                content, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )

    @method_decorator(api_staff_member_required())
    def post(self, request):
        """Create an elasticsearch template configuration

        Parameters:
            {
                "template": "template_id",
                "title_path": "/x/path/title",
                "description_paths": ["/x/path/desc_one", "/x/path/desc_two"],
            }

        Args:
            request: HTTP request

        Returns:

            - code: 201
              content: Created elasticsearch template
            - code: 400
              content: Validation error / not unique / model error
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializer
            serializer = ElasticsearchTemplateSerializer(data=request.data)

            # Validate data
            serializer.is_valid(raise_exception=True)

            # Save data
            serializer.save()

            return Response(serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as validation_exception:
            content = {"message": validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except exceptions.ModelError as model_exception:
            content = {"message": str(model_exception)}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except exceptions.NotUniqueError as not_unique_error:
            content = {"message": str(not_unique_error)}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except exceptions.ApiError as api_exception:
            content = {"message": str(api_exception)}
            return Response(
                content, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )
        except Exception as exception:
            content = {"message": str(exception)}
            return Response(
                content, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )



""" Elasticsearch client for the CDCS
"""
from elasticsearch import Elasticsearch

from core_elasticsearch_app.settings import (
    ELASTICSEARCH_HOST,
    ELASTICSEARCH_PORT,
)


class ElasticsearchClient:
    """Elasticsearch client"""

    _elasticsearch = None

    @classmethod
    def elasticsearch(cls):
        """Return Elasticsearch connection
        Returns:
        """
        if cls._elasticsearch is None:
            cls._elasticsearch = Elasticsearch(
                [{"host": ELASTICSEARCH_HOST, "port": ELASTICSEARCH_PORT}]
            )
        return cls._elasticsearch

    @classmethod
    def delete_index(cls, index_name):
        """delete_index
        Args:
            index_name:

        Returns:
        """
        return cls.elasticsearch().indices.delete(
            index=index_name, ignore=[400, 404]
        )

    @classmethod
    def create_index(cls, index_name, settings=None):
        """create_index
        Args:
            index_name:
            settings:

        Returns:
        """
        return cls.elasticsearch().indices.create(
            index=index_name, body=settings, ignore=400
        )

    @classmethod
    def index_document(cls, index_name, document_id, document):
        """index_document
        Args:
            index_name:
            document_id:
            document:

        Returns:
        """
        return cls.elasticsearch().index(
            index=index_name, id=document_id, body=document
        )

    @classmethod
    def delete_document(cls, index_name, document_id):
        """delete_document
        Args:
            index_name:
            document_id:

        Returns:
        """
        return cls.elasticsearch().delete(index=index_name, id=document_id)

    @classmethod
    def update_document(cls, index_name, document_id, document):
        """update_document
        Args:
            index_name:
            document_id:
            document:

        Returns:
        """
        return cls.elasticsearch().update(
            index=index_name, id=document_id, body=document
        )

    @classmethod
    def search(cls, index_name, query):
        """search
        Args:
            index_name:
            query:

        Returns:
        """
        return cls.elasticsearch().search(index=index_name, body=query)



""" Elasticsearch Ajax
"""
import json

from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse
from django.urls import reverse_lazy

from core_main_app.components.data import api as data_api
from core_main_app.views.common.ajax import (
    AddObjectModalView,
    DeleteObjectModalView,
    EditObjectModalView,
)

from core_elasticsearch_app.components.data.mongodb import (
    get_exists_query_from_path,
)
from core_elasticsearch_app.components.elasticsearch_template import (
    api as elasticsearch_template_api,
)
from core_elasticsearch_app.components.elasticsearch_template.models import (
    ElasticsearchTemplate,
)
from core_elasticsearch_app.tasks import index_all_data_from_template
from core_elasticsearch_app.views.admin.forms import ElasticsearchTemplateForm


class AddElasticsearchTemplateView(AddObjectModalView):
    """AddElasticsearchTemplateView"""

    form_class = ElasticsearchTemplateForm
    model = ElasticsearchTemplate
    success_url = reverse_lazy("core-admin:core_elasticsearch_app_templates")
    success_message = "Template was successfully configured for Elasticsearch."

    def _save(self, form):
        # Save treatment.
        try:
            es_template = ElasticsearchTemplate(
                template=form.cleaned_data["template"],
                title_path=form.cleaned_data["title_path"],
                description_paths=form.cleaned_data["description_paths"],
            )
            elasticsearch_template_api.upsert(es_template)
        except Exception as exception:
            form.add_error(None, str(exception))

    def get_form_kwargs(self, *args, **kwargs):
        """get_form_kwargs

        Returns:

        """
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["request"] = self.request
        return kwargs


class EditElasticsearchTemplateView(EditObjectModalView):
    """EditElasticsearchTemplateView"""

    form_class = ElasticsearchTemplateForm
    model = ElasticsearchTemplate
    success_url = reverse_lazy("core-admin:core_elasticsearch_app_templates")
    success_message = "Template Configuration edited with success."

    def _save(self, form):
        # Save treatment.
        try:
            self.object.description_paths = form.cleaned_data[
                "description_paths"
            ]
            elasticsearch_template_api.upsert(self.object)
        except Exception as exception:
            form.add_error(None, str(exception))

    def get_form_kwargs(self, *args, **kwargs):
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["request"] = self.request
        return kwargs


class DeleteElasticsearchTemplateView(DeleteObjectModalView):
    """DeleteElasticsearchTemplateView"""

    model = ElasticsearchTemplate
    success_url = reverse_lazy("core-admin:core_elasticsearch_app_templates")
    success_message = "Template Configuration deleted with success."

    def _delete(self, request, *args, **kwargs):
        # Delete treatment.
        elasticsearch_template_api.delete(self.object)

    def _get_object_name(self):
        return f"the configuration using the template {self.object.template.display_name}"

    def get_form_kwargs(self, *args, **kwargs):
        """get_form_kwargs

        Returns:

        """
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["request"] = self.request
        return kwargs


@staff_member_required
def check_data_from_template(request, pk):
    """Check if data found for provided xpaths

    Args:
        request:
        pk:

    Returns:

    """
    try:
        # Get elasticsearch configuration for the given template
        es_template = elasticsearch_template_api.get_by_id(pk)
        # Check if titles found with given path
        title_results = data_api.execute_json_query(
            json_query=get_exists_query_from_path(es_template.title_path),
            user=request.user,
        )
        # Check if descriptions found with given path
        desc_count = 0
        for path in es_template.description_paths:
            description_results = data_api.execute_json_query(
                json_query=get_exists_query_from_path(path),
                user=request.user,
            )
            desc_count += description_results.count()
        message = (
            f"Title: {title_results.count()} data found. "
            f"Description: {desc_count} data fields found."
        )
    except Exception as exception:
        message = str(exception)

    return HttpResponse(
        json.dumps(message), content_type="application/javascript"
    )


@staff_member_required
def index_data_from_template(request, pk):
    """Index all data from template

    Args:
        request:
        pk:

    Returns:

    """
    # Get elasticsearch configuration for the given template
    es_template = elasticsearch_template_api.get_by_id(pk)
    # Index all data for this template
    index_all_data_from_template.apply_async((str(es_template.template.id),))

    return HttpResponse(
        json.dumps("Indexing data..."), content_type="application/javascript"
    )



""" Elasticsearch forms
"""
import logging

from django import forms
from django.forms import ModelForm

from core_main_app.commons import exceptions
from core_main_app.components.template import api as template_api
from core_main_app.components.template_version_manager import (
    api as template_version_manager_api,
)

from core_elasticsearch_app.components.elasticsearch_template.models import (
    ElasticsearchTemplate,
)

logger = logging.getLogger(__name__)


class ElasticsearchTemplateForm(ModelForm):
    """ElasticsearchTemplate form"""

    template = forms.ChoiceField(
        label="Template", widget=forms.Select(attrs={"class": "form-control"})
    )

    title_path = forms.CharField(
        label="Path to Title",
        required=False,
        widget=forms.TextInput(
            attrs={
                "placeholder": "Resource/identity/title",
                "class": "form-control",
            }
        ),
    )

    description_paths = forms.CharField(
        label="Paths to Description",
        required=False,
        widget=forms.Textarea(
            attrs={
                "placeholder": "Resource/content/description\nResource/content/subject",
                "class": "form-control",
            }
        ),
    )

    class Meta:
        """Meta"""

        model = ElasticsearchTemplate
        fields = ["template", "title_path"]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)
        self.fields["template"].choices = _get_templates_versions(
            request=self.request
        )
        self.fields["description_paths"].initial = "\n".join(
            kwargs["instance"].description_paths if kwargs["instance"] else ""
        )

    def clean_template(self):
        """clean_template

        Returns:

        """
        data = self.cleaned_data["template"]
        return template_api.get_by_id(data, request=self.request)

    def clean_description_paths(self):
        """clean_description_paths

        Returns:

        """
        return self.cleaned_data["description_paths"].split()


# FIXME: duplicate from core_oaipmh_provider_app
def _get_templates_versions(request):
    """Get templates versions.

    Args:
        request:

    Returns:
        List of templates versions.

    """
    templates = []
    try:
        list_ = template_version_manager_api.get_active_global_version_manager(
            request=request
        )
        for elt in list_:
            for version in elt.versions:
                template = template_api.get_by_id(version, request=request)
                version_name = template.display_name
                templates.append((version, version_name))
    except exceptions.DoesNotExist as exception:
        logger.warning(
            "_get_templates_versions threw an exception: %s", str(exception)
        )

    return templates



""" Utils for the elasticsearch package
"""

import re
from functools import reduce


def clean_keyword(keyword):
    """Remove extra characters and format keyword

    Args:
        keyword:

    Returns:

    """
    # lowercase
    keyword = keyword.lower()

    # remove extra characters
    regex = re.compile("<em>.*</em>")
    keyword = re.search(regex, keyword).group()

    # remove highlight tags
    keyword = keyword.replace("<em>", "").replace("</em>", "")

    return keyword


def get_nested_value(dictionary, keys, default=None):
    """Get value in nested dictionary

    From https://stackoverflow.com/a/46890853

    Args:
        dictionary:
        keys:
        default:

    Returns:

    """
    return reduce(
        lambda d, key: d.get(key, default) if isinstance(d, dict) else default,
        keys.split("/"),
        dictionary,
    )



""" Migration
"""
# Generated by Django 3.2.10 on 2022-01-28 15:56

import core_main_app.utils.validation.xpath_validation
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    """Migration"""

    initial = True

    dependencies = [
        ("core_main_app", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ElasticsearchTemplate",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title_path", models.CharField(default=None, max_length=400)),
                (
                    "description_paths",
                    models.JSONField(
                        blank=True,
                        default=None,
                        null=True,
                        validators=[
                            core_main_app.utils.validation.xpath_validation.validate_xpath_list
                        ],
                    ),
                ),
                (
                    "template",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core_main_app.template",
                    ),
                ),
            ],
        ),
    ]



""" Admin views Core Elasticsearch Search App
"""
from django.views.generic import View


from core_main_app.utils.rendering import admin_render
from core_main_app.views.common.ajax import (
    AddObjectModalView,
    DeleteObjectModalView,
    EditObjectModalView,
)
from core_elasticsearch_app.components.elasticsearch_template import (
    api as elasticsearch_template_api,
)


class TemplatesView(View):
    """Templates View"""

    def get(self, request, *args, **kwargs):
        """Configure templates, Display as list.

        Args:
            request:

        Returns:

        """
        context = {
            "object_name": "Template Configuration",
            "es_template_list": elasticsearch_template_api.get_all(),
        }

        modals = [
            DeleteObjectModalView.get_modal_html_path(),
            AddObjectModalView.get_modal_html_path(),
            EditObjectModalView.get_modal_html_path(),
        ]

        assets = {
            "js": [
                {
                    "path": "core_elasticsearch_app/js/admin/es_template.js",
                    "is_raw": False,
                },
                DeleteObjectModalView.get_modal_js_path(),
                AddObjectModalView.get_modal_js_path(),
                EditObjectModalView.get_modal_js_path(),
            ]
        }

        return admin_render(
            request,
            "core_elasticsearch_app/admin/templates/list_templates.html",
            assets=assets,
            context=context,
            modals=modals,
        )



""" MongoDB utils for the Data component
"""
import logging

from core_main_app.utils.xml import xpath_to_dot_notation
from core_elasticsearch_app.utils.utils import get_nested_value

logger = logging.getLogger(__name__)


def get_value_from_path(data, path):
    """Get value from xpath

    Args:
        data:
        path:

    Returns:

    """
    value = get_nested_value(data.get_dict_content(), path)
    if value:
        if isinstance(value, dict):
            logger.info(
                "A dict was found at %s for data %s ", path, str(data.id)
            )
            try:
                value = value["#text"]
            except Exception:
                value = None
                logger.warning(
                    "#text was not found at path %s for data %s ",
                    path,
                    str(data.id),
                )
    else:
        logger.warning(
            "No value could be found at path %s for data %s",
            path,
            str(data.id),
        )
    return value


def get_exists_query_from_path(path):
    """Return mongodb query that checks if path exists in database

    Args:
        path:

    Returns:

    """
    dot_path = xpath_to_dot_notation(path)
    return {
        "$or": [
            {f"dict_content.{dot_path}": {"$exists": True}},
            {f"dict_content.{dot_path}.#text": {"$exists": True}},
        ]
    }



""" Watchers for the data collection
"""
from django.db.models.signals import post_save

from core_main_app.components.data.models import Data
from core_elasticsearch_app.settings import ELASTICSEARCH_AUTO_INDEX
from core_elasticsearch_app.tasks import index_data


def post_save_data(sender, instance, **kwargs):
    """Method executed after a saving of a Data object.
    Args:
        sender: Class.
        instance: Data document.
        **kwargs: Args.

    """
    # only deal with data in a workspace
    if not instance.workspace:
        return

    # only deal with public data
    from core_main_app.components.workspace import api as workspace_api

    public_workspaces = workspace_api.get_all_public_workspaces().values_list(
        "id", flat=True
    )
    if instance.workspace.id not in public_workspaces:
        return

    # Creates or updates a document in the index.
    index_data.apply_async((str(instance.id),))


def init():
    """Connect to Data object events."""
    if ELASTICSEARCH_AUTO_INDEX:
        post_save.connect(post_save_data, sender=Data)



""" ElasticsearchTemplate model
"""
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from core_main_app.commons import exceptions
from core_main_app.components.template.models import Template
from core_main_app.utils.validation.xpath_validation import validate_xpath_list


class ElasticsearchTemplate(models.Model):
    """ElasticsearchTemplate object"""

    template = models.OneToOneField(
        Template, blank=False, on_delete=models.CASCADE, unique=True
    )
    # optional - default data.title
    title_path = models.CharField(default=None, max_length=400)
    # optional - default to data full text
    description_paths = models.JSONField(
        blank=True, null=True, validators=[validate_xpath_list], default=None
    )

    @staticmethod
    def get_by_id(es_template_id):
        """Returns the object with a given id

        Args:
            es_template_id:

        Returns:
            ElasticsearchTemplate (obj): ElasticsearchTemplate

        """
        try:
            return ElasticsearchTemplate.objects.get(pk=es_template_id)
        except ObjectDoesNotExist as exception:
            raise exceptions.DoesNotExist(str(exception))
        except Exception as exception:
            raise exceptions.ModelError(str(exception))

    @staticmethod
    def get_by_template(template):
        """Returns the object with a given template

        Args:
            template:

        Returns:
            ElasticsearchTemplate (obj): ElasticsearchTemplate

        """
        try:
            return ElasticsearchTemplate.objects.get(template=template)
        except ObjectDoesNotExist as exception:
            raise exceptions.DoesNotExist(str(exception))
        except Exception as exception:
            raise exceptions.ModelError(str(exception))

    @staticmethod
    def get_all():
        """Returns all objects"""
        return ElasticsearchTemplate.objects.all()

    def __str__(self):
        """Elasticsearch template object as string

        Returns:

        """
        return self.template.__str__()



""" REST views for the elasticsearch package
"""

from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from core_elasticsearch_app.components.data.elasticsearch import (
    get_suggestions,
)
from core_elasticsearch_app.utils.utils import clean_keyword


class DocumentSuggestion(APIView):
    """Get document suggestions from keywords."""

    def get(self, request):
        """Get all Suggestions

        Url Parameters:

            keywords: keywords

        Examples:

            ../suggestions?keywords=[keywords]

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of suggestions
            - code: 500
              content: Internal server error
        """
        try:
            # Get objects
            keywords = self.request.query_params.get("keywords", [])
            if keywords:
                keywords = keywords.split(",")
            suggestions = get_suggestions(" ".join(keywords))
            return Response(suggestions, status=status.HTTP_200_OK)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(
                content, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )


class KeywordSuggestion(APIView):
    """Get Keyword suggestions from keywords."""

    def get(self, request):
        """Get all Suggestions

        Url Parameters:

            keywords: keywords

        Examples:

            ../suggestions?keywords=[keywords]

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of keywords
            - code: 400
              content: Bad Request
            - code: 500
              content: Internal server error
        """
        try:
            # Get keyword parameter
            keyword = self.request.query_params.get("keyword", None)
            if not keyword:
                content = {"message": "Keyword parameter is missing"}
                return Response(content, status=status.HTTP_400_BAD_REQUEST)

            # get suggestions from elasticsearch
            suggestions = get_suggestions(keyword, fragment_size=1)
            # build a list of formatted keywords
            suggestions = [
                clean_keyword(s["highlight"]["description"][0])
                for s in suggestions
                if "description" in s["highlight"]
            ]
            # get list of unique keywords
            suggestions = set(suggestions)
            return Response(suggestions, status=status.HTTP_200_OK)

        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(
                content, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )



""" Settings for core_elasticsearch_app.

Settings with the following syntax can be overwritten at the project level:
SETTING_NAME = getattr(settings, "SETTING_NAME", "Default Value")
"""
from django.conf import settings

if not settings.configured:
    settings.configure()

ELASTICSEARCH_HOST = getattr(settings, "ELASTICSEARCH_HOST", "localhost")
""" :py:class:`str`: Elasticsearch host
"""

ELASTICSEARCH_PORT = getattr(settings, "ELASTICSEARCH_PORT", 9200)
""" :py:class:`int`: Elasticsearch port
"""

CAN_SET_PUBLIC_DATA_TO_PRIVATE = getattr(
    settings, "CAN_SET_PUBLIC_DATA_TO_PRIVATE", False
)
""" :py:class:`boolean`: True if public data can be unpublished
"""

CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT = getattr(
    settings, "CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT", False
)
""" :py:class:`bool`: Can anonymous user access public document.
"""

ELASTICSEARCH_CDCS_DATA_INDEX = getattr(
    settings, "ELASTICSEARCH_CDCS_DATA_INDEX", "cdcs-data"
)
""" :py:class:`str`: Name of the Elasticsearch index for CDCS data
"""

ELASTICSEARCH_AUTO_INDEX = getattr(settings, "ELASTICSEARCH_AUTO_INDEX", True)
""" :py:class:`boolean`: True if data should be automatically indexed in Elasticsearch
"""
